from injectable import injectable
import random
from string import ascii_lowercase, ascii_uppercase
from user_agent import generate_user_agent


@injectable
class UtilsRandom:
    def __init__(self):
        self.CHARS_ASCII = []
        self.CHARS_ASCII.extend(ascii_lowercase)
        self.CHARS_ASCII.extend(ascii_uppercase)

    def get_random_string(self, size: int) -> str:
        return ''.join([random.choice(self.CHARS_ASCII) for i in range(0, size)])

    def get_random_user_agent(self) -> str:
        return str(generate_user_agent())

    def get_random_ip(self) -> str:
        return '.'.join([str(random.randint(0,255)) for i in range(0,4)])



from datetime import datetime, tzinfo
from typing import List

import pytz
import tzlocal
import calendar
from injectable import injectable

DEFAULT_TZ = pytz.utc
DEFAULT_DATE_FORMAT = '%d.%m.%Y %H:%M:%S (%z)'


@injectable
class UtilsTime:
    # Timezones
    #-------------------------------------------------------------------------------------------------------------------
    def get_timezone(self, name: str) -> tzinfo:
        return pytz.timezone(name)

    def get_timezone_local(self) -> tzinfo:
        return tzlocal.get_localzone()

    def get_timezone_utc(self) -> tzinfo:
        return self.get_timezone("UTC")

    def get_timezone_pst(self) -> tzinfo:
        return self.get_timezone("US/Pacific")
    #-------------------------------------------------------------------------------------------------------------------



    # Timestamps
    #-------------------------------------------------------------------------------------------------------------------
    def get_timestamp_ms_now(self, timezone: tzinfo = DEFAULT_TZ) -> int:
        date_time_object = datetime.now(tz=timezone)
        timestamp_s = date_time_object.timestamp()
        return int(round(timestamp_s * 1000))

    def get_timestamp_ms_day(self, year: int, month: int, day: int, timezone: tzinfo = DEFAULT_TZ) -> int:
        date_time_object = datetime(year=year, month=month, day=day, hour=12, minute=0, second=0).astimezone(timezone)
        timestamp_s = date_time_object.timestamp()
        return int(round(timestamp_s * 1000))
    #-------------------------------------------------------------------------------------------------------------------



    # Formating to string / parsing from string
    #-------------------------------------------------------------------------------------------------------------------
    def format_timestamp_ms(self, timestamp_ms: int, timezone: tzinfo = DEFAULT_TZ, date_format: str = DEFAULT_DATE_FORMAT) -> str:
        date_time_object = datetime.fromtimestamp(timestamp_ms / 1000, tz=timezone)
        return date_time_object.strftime(date_format)

    def parse_timestamp_ms(self, date_str: str, date_format: str = DEFAULT_DATE_FORMAT) -> int:
        date_time_object = datetime.strptime(date_str, date_format)
        timestamp_s = date_time_object.timestamp()
        return int(round(timestamp_s * 1000))
    #-------------------------------------------------------------------------------------------------------------------



    # Day
    #-------------------------------------------------------------------------------------------------------------------
    def get_day_start_timestamp_ms(self, year: int, month: int, day: int, timezone: tzinfo = DEFAULT_TZ) -> int:
        date_time_object = datetime(year=year, month=month, day=day, hour=0, minute=0, second=1).astimezone(timezone)
        timestamp_s = date_time_object.timestamp()
        return int(round(timestamp_s * 1000))

    def get_day_end_timestamp_ms(self, year: int, month: int, day: int, timezone: tzinfo = DEFAULT_TZ) -> int:
        date_time_object = datetime(year=year, month=month, day=day, hour=23, minute=59, second=59).astimezone(timezone)
        timestamp_s = date_time_object.timestamp()
        return int(round(timestamp_s * 1000))

    def get_day_start_end_timestamps_ms(self, year: int, month: int, day: int, timezone: tzinfo = DEFAULT_TZ) -> (int, int):
        return self.get_day_start_timestamp_ms(year, month, day, timezone=timezone), self.get_day_end_timestamp_ms(year, month, day, timezone=timezone)
    #-------------------------------------------------------------------------------------------------------------------



    # Month
    #-------------------------------------------------------------------------------------------------------------------
    def get_month_number_of_days(self, year: int, month: int) -> int:
        x, number_of_days_in_month = calendar.monthrange(year, month)
        return number_of_days_in_month

    def get_month_start_timestamp_ms(self, year: int, month: int, timezone: tzinfo = DEFAULT_TZ) -> int:
        return self.get_day_start_timestamp_ms(year, month, 1, timezone=timezone)

    def get_month_end_timestamp_ms(self, year: int, month: int, timezone: tzinfo = DEFAULT_TZ) -> int:
        number_of_days_in_month = self.get_month_number_of_days(year, month)
        return self.get_day_end_timestamp_ms(year, month, number_of_days_in_month, timezone=timezone)

    def get_month_start_end_timestamps_ms(self, year: int, month: int, timezone: tzinfo = DEFAULT_TZ) -> (int, int):
        return self.get_month_start_timestamp_ms(year, month, timezone=timezone), self.get_month_end_timestamp_ms(year, month, timezone=timezone)

    def get_month_name(self, month: int) -> str:
        return calendar.month_name[month]

    def is_timestamp_in_month(self, timestamp_ms: int, year: int, month: int, timezone: tzinfo = DEFAULT_TZ) -> bool:
        month_start_ms, month_end_ms = self.get_month_start_end_timestamps_ms(year, month, timezone=timezone)
        return month_start_ms <= timestamp_ms <= month_end_ms

    def is_timestamp_in_months(self, timestamp_ms: int, year: int, months: List[int], timezone: tzinfo = DEFAULT_TZ) -> bool:
        for month in months:
            if self.is_timestamp_in_month(timestamp_ms, year, month, timezone=timezone):
                return True
        return False
    #-------------------------------------------------------------------------------------------------------------------



    # Year
    #-------------------------------------------------------------------------------------------------------------------
    def get_year_start_timestamp_ms(self, year: int, timezone: tzinfo = DEFAULT_TZ) -> int:
        return self.get_month_start_timestamp_ms(year, 1, timezone=timezone)

    def get_year_end_timestamp_ms(self, year: int, timezone: tzinfo = DEFAULT_TZ) -> int:
        return self.get_month_end_timestamp_ms(year, 12, timezone=timezone)

    def get_year_start_end_timestamps_ms(self, year: int, timezone: tzinfo = DEFAULT_TZ) -> (int, int):
        return self.get_year_start_timestamp_ms(year, timezone=timezone), self.get_year_end_timestamp_ms(year, timezone=timezone)

    def is_timestamp_in_year(self, timestamp_ms: int, year: int, timezone: tzinfo = DEFAULT_TZ) -> bool:
        year_start_ms, year_end_ms = self.get_year_start_end_timestamps_ms(year, timezone=timezone)
        return year_start_ms <= timestamp_ms <= year_end_ms
    #-------------------------------------------------------------------------------------------------------------------



import os
import threading
from typing import List
from injectable import injectable

environment_variables_map = {}
environment_variables_map_lock = threading.Lock()


# A utils class that provides common methods to deal with global environment variables
# This class has 2 main functions - load variables into memory, retrieve variables from memory
# To set these variables use docker's "-e" flag
#
# Try to follow the convention, variable name can be like "CASEDRIVE_ENV" or "A360_ENV" or "A360_SPACY_VERSION"
# And values can be like "local", "test", "prod", "v1", "default" etc
@injectable
class UtilsEnv:
    def load_environment_variable(self, environment_variable_name: str) -> bool:
        with environment_variables_map_lock:
            if environment_variable_name in os.environ:
                environment_variables_map[environment_variable_name] = os.environ[environment_variable_name]
                return True
            else:
                print('UtilsEnv.load_environment_variable(): Warning! Environment variable "' + str(environment_variable_name) + '"  not found')
                return False

    def load_environment_variables(self, list_of_environment_variable_names: List[str]) -> List[bool]:
        results = []
        for environment_variable_name in list_of_environment_variable_names:
            result = self.load_environment_variable(environment_variable_name)
            results.append(result)
        return results

    def get_environment_variable(self, environment_variable_name: str, default_value: str = '') -> str:
        if environment_variable_name in environment_variables_map:
            return environment_variables_map[environment_variable_name]
        else:
            return default_value



import cv2
import numpy
from numpy import ndarray
from injectable import injectable


@injectable
class UtilsOpencv:
    def get_dimensions_hw(self, image_cv: ndarray) -> (int, int):
        h, w = image_cv.shape[:2]
        return h, w

    def get_dimensions_wh(self, image_cv: ndarray) -> (int, int):
        h, w = image_cv.shape[:2]
        return w, h

    def get_number_of_channels(self, image_cv: ndarray) -> int:
        if image_cv.ndim == 2:
            return 1
        elif image_cv.ndim == 3:
            return image_cv.shape[-1]
        else:
            raise ValueError("Weird image with ndim=" + str(image_cv.ndim))

    def are_images_equal(self, image_cv_a: ndarray, image_cv_b: ndarray) -> bool:
        if image_cv_a.shape == image_cv_b.shape:
            difference = cv2.subtract(image_cv_a, image_cv_b)
            b, g, r = cv2.split(difference)
            if cv2.countNonZero(b) == 0 and cv2.countNonZero(g) == 0 and cv2.countNonZero(r) == 0:
                return True
            else:
                return False
        else:
            return False

    def convert_to_grayscale(self, image_cv: ndarray) -> ndarray:
        if self.get_number_of_channels(image_cv) == 1:
            return image_cv
        else:
            return cv2.cvtColor(image_cv, cv2.COLOR_BGR2GRAY)

    def blur_gaussian(self, image_cv: ndarray, blur_x: int, blur_y: int) -> ndarray:
        new_image_cv = image_cv.copy()
        new_image_cv = cv2.GaussianBlur(new_image_cv, (blur_x, blur_y), 0)
        return new_image_cv

    # Image transformations
    #-------------------------------------------------------------------------------------------------------------------
    # Rotate (preserving original dimensions of the image)
    def rotate_bound(self, image_cv: ndarray, angle: float) -> ndarray:
        if angle == 0:
            return image_cv
        image_result = image_cv.copy()
        h, w = image_result.shape[:2]
        c_x, c_y, = (w // 2, h // 2)
        rotation_matrix = cv2.getRotationMatrix2D((c_x, c_y), angle, 1.0)
        image_result = cv2.warpAffine(image_result, rotation_matrix, (w, h), flags=cv2.INTER_CUBIC, borderMode=cv2.BORDER_REPLICATE)
        return image_result

    # Rotate (changing dimensions of the original image)
    def rotate_free(self, image_cv: ndarray, angle: float) -> ndarray:
        if angle == 0:
            return image_cv
        image_result = image_cv.copy()
        h, w = image_result.shape[:2]
        c_x, c_y, = (w // 2, h // 2)

        # Compute the rotation matrix
        rotation_matrix = cv2.getRotationMatrix2D((c_x, c_y), angle, 1.0)
        rotation_matrix_cos = numpy.abs(rotation_matrix[0, 0])
        rotation_matrix_sin = numpy.abs(rotation_matrix[0, 1])

        # Compute the new bounding dimensions of the image
        new_w = int((h * rotation_matrix_sin) + (w * rotation_matrix_cos))
        new_h = int((h * rotation_matrix_cos) + (w * rotation_matrix_sin))

        # Adjust the rotation matrix to take into account translation
        rotation_matrix[0, 2] += (new_w / 2) - c_x
        rotation_matrix[1, 2] += (new_h / 2) - c_y

        # Rotate
        image_result = cv2.warpAffine(image_result, rotation_matrix, (new_w, new_h), flags=cv2.INTER_CUBIC, borderMode=cv2.BORDER_REPLICATE)
        return image_result

    # Rotate (smartly, deciding which way of rotation will work better here)
    def rotate(self, image_cv: ndarray, angle: float, rotate_bound_threshold_angle: float = 20.0) -> ndarray:
        if -1.0 * rotate_bound_threshold_angle <= angle <= rotate_bound_threshold_angle:
            return self.rotate_bound(image_cv, angle)
        else:
            return self.rotate_free(image_cv, angle)
    #-------------------------------------------------------------------------------------------------------------------



    # Image deskewing
    #-------------------------------------------------------------------------------------------------------------------
    # Calculate skew angle of an image
    def calculate_skew_angle(self, image_cv: ndarray) -> float:
        # Prep image, copy, convert to gray scale, blur, and threshold
        image_result = image_cv.copy()
        gray = self.convert_to_grayscale(image_result)
        blur = self.blur_gaussian(gray, 9, 9)
        thresh = cv2.threshold(blur, 0, 255, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU)[1]

        # Apply dilate to merge text into meaningful lines/paragraphs.
        # Use larger kernel on X axis to merge characters into single line, cancelling out any spaces.
        # But use smaller kernel on Y axis to separate between different blocks of text
        kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (30, 5))
        dilate = cv2.dilate(thresh, kernel, iterations=5)

        # Find all contours
        contours, hierarchy = cv2.findContours(dilate, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
        contours = sorted(contours, key = cv2.contourArea, reverse = True)
        if len(contours) == 0:
            return 0

        # Find largest contour and surround in min area box
        largestContour = contours[0]
        minAreaRect = cv2.minAreaRect(largestContour)

        # Determine the angle. Convert it to the value that was originally used to obtain skewed image
        angle = minAreaRect[-1]
        if angle < -45:
            angle = 90 + angle
            return -1.0 * angle
        elif angle > 45:
            angle = 90 - angle
            return angle
        return -1.0 * angle

    # Deskew image
    def deskew(self, image_cv: ndarray) -> (ndarray, float):
        angle = self.calculate_skew_angle(image_cv)
        rotated_image_cv = self.rotate(image_cv, -1.0 * angle)
        return rotated_image_cv, angle
    #-------------------------------------------------------------------------------------------------------------------



import os
from configparser import ConfigParser, SectionProxy
import threading
from injectable import injectable

config_files_map = {}
config_files_map_lock = threading.Lock()


# A utils class that provides common methods to work with configuration files
@injectable
class UtilsConfig:
    def load_config(self, config_file_path: str) -> bool:
        if os.path.exists(config_file_path):
            with config_files_map_lock:
                config = ConfigParser()
                config.read(config_file_path)
                config_files_map[config_file_path] = config
                return True
        else:
            raise RuntimeError('File at path "' + str(config_file_path) + '" does not exist')

    def get_config(self, config_file_path: str) -> ConfigParser:
        if config_file_path in config_files_map:
            return config_files_map[config_file_path]
        else:
            raise RuntimeError('Config at path "' + str(config_file_path) + '" was not loaded')

    def get_config_section(self, config_file_path: str, section: str) -> SectionProxy:
        config = self.get_config(config_file_path)
        if section in config:
            return config[section]
        else:
            raise RuntimeError('Config at path "' + str(config_file_path) + '" does not have section "' + str(section) + '"')

    def get_config_section_value(self, config_file_path: str, section: str, key: str) -> str:
        section = self.get_config_section(config_file_path, section)
        if key in section:
            return section[key]
        else:
            raise RuntimeError('Config at path "' + str(config_file_path) + '" in section "' + str(section) + '" does not have key "' + str(key) + '"')



from .utils_class import UtilsClass
from .utils_config import UtilsConfig
from .utils_env import UtilsEnv
from .utils_id import UtilsId
from .utils_image import UtilsImage
from .utils_opencv import UtilsOpencv
from .utils_ping import UtilsPing
from .utils_random import UtilsRandom
from .utils_string import UtilsString
from .utils_time import UtilsTime

__all__ = [
    UtilsClass,
    UtilsConfig,
    UtilsEnv,
    UtilsId,
    UtilsImage,
    UtilsOpencv,
    UtilsPing,
    UtilsRandom,
    UtilsString,
    UtilsTime
]



from injectable import injectable, autowired, Autowired
from tekleo_common_message_protocol import PingOutput
from tekleo_common_utils.utils_id import UtilsId
from tekleo_common_utils.utils_time import UtilsTime


@injectable
class UtilsPing:
    @autowired
    def __init__(self, utils_id: Autowired(UtilsId), utils_time: Autowired(UtilsTime)):
        self.utils_id = utils_id
        self.utils_time = utils_time

    def build(self) -> PingOutput:
        id = self.utils_id.generate_uuid()
        success = True
        timestamp = self.utils_time.get_timestamp_ms_now()
        return PingOutput(id, success, timestamp)



from PIL import Image as image_pil_main
from PIL.Image import Image
import cv2
import numpy
import requests
import tempfile
import io
import base64
from numpy import ndarray
from injectable import injectable


@injectable
class UtilsImage:
    def convert_image_pil_to_image_cv(self, image_pil: Image) -> ndarray:
        return cv2.cvtColor(numpy.array(image_pil), cv2.COLOR_RGB2BGR)

    def convert_image_cv_to_image_pil(self, image_cv: ndarray) -> Image:
        return image_pil_main.fromarray(cv2.cvtColor(image_cv, cv2.COLOR_BGR2RGB))

    def open_image_pil(self, image_path: str) -> Image:
        return image_pil_main.open(image_path)

    def open_image_cv(self, image_path: str) -> ndarray:
        return self.convert_image_pil_to_image_cv(self.open_image_pil(image_path))

    def save_image_pil(self, image_pil: Image, image_path: str) -> str:
        image_pil.save(image_path)
        return image_path

    def debug_image_cv(self, image_cv: ndarray, window_name: str = 'Debug Image'):
        cv2.namedWindow(window_name, cv2.WINDOW_NORMAL)
        cv2.imshow(window_name, image_cv)
        cv2.waitKey(0)
        cv2.destroyAllWindows()

    def download_image_pil(self, image_url: str, timeout_in_seconds: int = 90) -> Image:
        # Make request
        response = requests.get(image_url, timeout=timeout_in_seconds, stream=True)
        response.raise_for_status()

        # Download the image into buffer
        buffer = tempfile.SpooledTemporaryFile(max_size=1e9)
        downloaded = 0
        for chunk in response.iter_content(chunk_size=1024):
            downloaded += len(chunk)
            buffer.write(chunk)
        buffer.seek(0)

        # Convert buffer to image
        image = image_pil_main.open(io.BytesIO(buffer.read()))
        return image

    def encode_image_pil_as_base64(self, image_pil: Image) -> str:
        bytes_io = io.BytesIO()
        image_pil.save(bytes_io, format="PNG")
        return str(base64.b64encode(bytes_io.getvalue()), 'utf-8')

    def decode_image_pil_from_base64(self, image_base64: str) -> Image:
        image_bytes = base64.b64decode(bytes(image_base64, 'utf-8'))
        image = image_pil_main.open(io.BytesIO(image_bytes))
        return image



from injectable import injectable


@injectable
class UtilsString:
    def remove_mixed_spaces(self, text: str) -> str:
        words = text.split(' ')
        new_words = [word.strip() for word in words if len(word.strip()) > 0]
        return ' '.join(new_words)
